import pytest
import io
import pandas as pd
import random

from datetime import timedelta
from tqdm import tqdm
from pathlib import Path
from tempfile import NamedTemporaryFile
from unittest import mock

from ketl.db import models
from ketl.extractor.Extractor import SourceTargetPair, BaseExtractor, DefaultExtractor
from ketl.utils.file_utils import file_hash
from tests.factories import APIFactory, CredsFactory, SourceFactory, CachedFileFactory, ExpectedFileFactory


def test_base_extractor():

    extractor = BaseExtractor()
    with pytest.raises(NotImplementedError):

        extractor.extract()


def test_default_extractor_init():

    api: models.API = APIFactory(name='my nice api')
    creds_details = {
        'cookie': {
            'name': 'my-cookie',
            'value': 'my-value'
        },
        'auth': {
            'username': 'my-username',
            'password': 'my-password'
        },
        'auth_token': {
            'header': 'Token',
            'token': 'my-token'
        }
    }
    creds: models.Creds = CredsFactory(api_config=api, creds_details=creds_details)

    extractor = DefaultExtractor(api)

    assert extractor.auth == creds_details['auth']
    assert extractor.headers['Cookie'] == 'my-cookie=my-value'
    assert extractor.auth_token == creds_details['auth_token']
    assert extractor.headers == {
        'Cookie': 'my-cookie=my-value',
        'Token': 'my-token'
    }


def test_source_target_path():

    download_dir = '/path/to/download'
    source_url = 'http://path/to/source'
    api: models.API = APIFactory(name='my nice api')
    source: models.Source = SourceFactory(base_url=source_url, data_dir=download_dir, api_config=api)

    cached_file1: models.CachedFile = CachedFileFactory(url='path/to/file1',
                                                        path='path/to/file1/on/disk', source=source)
    cached_file2: models.CachedFile = CachedFileFactory(url='path/to/file2', source=source)

    extractor = DefaultExtractor(api)

    expected_source_targets = [
        SourceTargetPair(cached_file1, Path(download_dir).resolve() / 'path/to/file1/on/disk'),
        SourceTargetPair(cached_file2, Path(download_dir).resolve() / 'file2')
    ]

    assert expected_source_targets == extractor.source_target_list


def test_ftp_writer():

    target = io.BytesIO()

    bar = tqdm(1)
    DefaultExtractor._ftp_writer(target, b'hello world', bar=bar)
    target.seek(0)
    bar.close()

    assert target.read() == b'hello world'


def test_generic_writer():

    size = 32768
    bar = tqdm(size)
    letters = [chr(i) for i in range(65, 65 + 26)]
    data = bytes(''.join(random.choices(letters, k=size)), 'utf-8')

    source = io.BytesIO(data)
    source.seek(0)
    target = io.BytesIO()

    api: models.API = APIFactory(name='my nice api')
    extractor = DefaultExtractor(api)
    extractor._generic_writer(source, target, bar=bar)

    source.seek(0)
    target.seek(0)
    assert source.read() == target.read()


def test_requires_update(tmp_path):

    api: models.API = APIFactory(name='my nice api')
    extractor = DefaultExtractor(api)

    assert extractor._requires_update(Path('some/file'), 1, timedelta(days=1))

    tf = NamedTemporaryFile(dir=tmp_path, delete=False)
    tf.write(b'hello world')
    tf.close()

    assert not extractor._requires_update(Path(tf.name), 11)
    assert not extractor._requires_update(Path(tf.name), -1, timedelta(days=1))
    assert extractor._requires_update(Path(tf.name), -1, timedelta(0))


def test_update_cache_file(session, tmp_path):

    tf = NamedTemporaryFile(dir=tmp_path, delete=False)
    tf.write(b'hello world')
    tf.close()

    api: models.API = APIFactory(name='my nice api')
    extractor = DefaultExtractor(api)

    cached_file: models.CachedFile = CachedFileFactory(path=tf.name)
    extractor._update_file_cache(cached_file, Path(tf.name))

    cached_file = session.query(models.CachedFile).get(cached_file.id)

    assert cached_file.path == tf.name
    assert cached_file.hash == file_hash(Path(tf.name)).hexdigest()
    assert cached_file.last_download is not None
    assert cached_file.size == 11


@mock.patch('ketl.extractor.Extractor.DefaultExtractor.get_file')
@mock.patch('ketl.extractor.Extractor.DefaultExtractor.source_target_list', new_callable=mock.PropertyMock)
def test_extract(mock_source_files_list, mock_get_file):

    api: models.API = APIFactory()
    source: models.Source = SourceFactory(api_config=api, data_dir='data/dir')

    extractor = DefaultExtractor(api)

    cached_file1: models.CachedFile = CachedFileFactory(url='url/to/file1', path='path/to/file1',
                                                        expected_mode=models.ExpectedMode.self,
                                                        source=source)
    cached_file2: models.CachedFile = CachedFileFactory(url='url/to/file2', path='path/to/file2',
                                                        expected_mode=models.ExpectedMode.self,
                                                        source=source)
    cached_file3: models.CachedFile = CachedFileFactory(url='url/to/file3', path='path/to/file3',
                                                        expected_mode=models.ExpectedMode.self,
                                                        source=source)

    mock_source_files_list.return_value = [
        SourceTargetPair(cached_file1, Path('path/to/file1')),
        SourceTargetPair(cached_file2, Path('path/to/file2')),
        SourceTargetPair(cached_file3, Path('path/to/file3')),
    ]

    mock_get_file.side_effect = [cached_file1, cached_file2, None]

    expected_files = extractor.extract()

    for i, ef in enumerate(expected_files):
        assert str(ef) == f'data/dir/path/to/file{i + 1}'


@mock.patch('ketl.extractor.Extractor.FTP')
@mock.patch('ketl.extractor.Extractor.DefaultExtractor._requires_update')
def test_fetch_ftp(mock_requires_update, mock_ftp_class, tmp_path):

    mock_requires_update.return_value = True

    mock_ftp = mock.Mock()
    mock_ftp.login.return_value = None
    mock_ftp.size.return_value = 11
    mock_ftp.retrbinary.return_value = None

    mock_ftp_class.return_value = mock_ftp

    api: models.API = APIFactory(name='my nice api')
    extractor = DefaultExtractor(api)

    cached_file: models.CachedFile = CachedFileFactory(url='url/to/file')

    with NamedTemporaryFile(dir=tmp_path) as tf:
        result = extractor._fetch_ftp_file(cached_file, Path(tf.name), tqdm(1))

    assert result

    mock_requires_update.return_value = False
    assert not extractor._fetch_ftp_file(cached_file, Path(tf.name), tqdm(1))


@mock.patch('ketl.extractor.Extractor.smart_open')
@mock.patch('ketl.extractor.Extractor.DefaultExtractor._requires_update')
@mock.patch('ketl.extractor.Extractor.DefaultExtractor._generic_writer')
def test_fetch_generic_file(mock_generic_writer, mock_requires_update, mock_smart_open, tmp_path):

    mock_requires_update.return_value = True

    mock_open_file = mock.Mock()
    mock_open_file.content_length = 1

    mock_smart_open.return_value.__enter__.return_value = mock_open_file

    mock_generic_writer.return_value = None

    api: models.API = APIFactory(name='my nice api')
    extractor = DefaultExtractor(api)

    cached_file: models.CachedFile = CachedFileFactory(url='url/to/file', url_params={'query': 'item'})

    with NamedTemporaryFile(dir=tmp_path) as tf:
        result = extractor._fetch_generic_file(cached_file, Path(tf.name),
                                               headers={'Bearer': 'Token'},
                                               auth={'username': 'user', 'password': 'password'})

    assert result

    mock_requires_update.return_value = False
    assert not extractor._fetch_generic_file(cached_file, Path(tf.name))


@mock.patch('ketl.extractor.Extractor.DefaultExtractor._update_file_cache')
@mock.patch('ketl.extractor.Extractor.DefaultExtractor._fetch_generic_file')
@mock.patch('ketl.extractor.Extractor.DefaultExtractor._fetch_ftp_file')
def test_get_file(mock_fetch_ftp, mock_fetch_generic, mock_update):

    mock_update.return_value = None

    api = APIFactory(name='my nice api')
    source1: models.Source = SourceFactory(base_url='ftp://base/url')
    source2: models.Source = SourceFactory(base_url='http://base/url')
    cached_file: models.CachedFile = CachedFileFactory(url='file', source=source1, url_params={'query': 'item'})

    extractor = DefaultExtractor(api)

    mock_fetch_ftp.return_value = True
    assert extractor.get_file(cached_file, Path('target')) == cached_file

    cached_file: models.CachedFile = CachedFileFactory(url='http://url/to/file', source=source2,
                                                       url_params={'query': 'item'})

    mock_fetch_generic.return_value = True
    assert extractor.get_file(cached_file, Path('target')) == cached_file

    mock_fetch_generic.return_value = False
    assert extractor.get_file(cached_file, Path('target')) is None

    mock_fetch_generic.side_effect = [ValueError('something bad happened')]
    mock_fetch_generic.return_value = True
    assert extractor.get_file(cached_file, Path('target')) is None



import requests

from marshmallow import Schema
from typing import Optional
from urllib.parse import urljoin


# does this need to be separate from the model?

class RestMixin:

    def get(self, base_url, resource, params=None, data_schema: Schema = None, result_schema: Schema = None, **kwargs):

        url = urljoin(base_url, resource)
        return self._execute_request(url, 'GET', params=params, data_schema=data_schema,
                                     result_schema=result_schema, **kwargs)

    def post(self, base_url, resource, data=None, json=None, data_schema: Optional[Schema] = None,
             result_schema: Optional[Schema] = None, **kwargs):

        url = urljoin(base_url, resource)
        return self._execute_request(url, 'POST', data=data, json=json,
                                     data_schema=data_schema, result_schema=result_schema, **kwargs)

    def put(self, base_url, resource, data=None, json=None, data_schema: Optional[Schema] = None,
            result_schema: Optional[Schema] = None, **kwargs):

        url = urljoin(base_url, resource)
        return self._execute_request(url, 'PUT', data=data, json=json,
                                     data_schema=data_schema, result_schema=result_schema, **kwargs)

    @staticmethod
    def _execute_request(url, method: str, params=None, data=None, json=None,
                         data_schema: Optional[Schema] = None, result_schema: Optional[Schema] = None, **kwargs):

        if method.upper() == 'GET':
            if params and data_schema:
                params = data_schema.load(params)
            result = requests.get(url, params, **kwargs)
        elif method.upper() == 'POST':
            if data_schema and json:
                json = data_schema.load(json)
            result = requests.post(url, data=data, json=json, **kwargs)
        elif method.upper() == 'PUT':
            if data_schema and json:
                json = data_schema.load(json)
            result = requests.put(url, data=data, json=json, **kwargs)
        else:
            raise ValueError(f'Unsupported method: {method}')

        try:
            json_result = result.json()
        except:
            import ipdb
            ipdb.set_trace()
            
        if result_schema:
            return result_schema.load(data=json_result)
        else:
            return json_result



import factory
from factory.alchemy import SQLAlchemyModelFactory

from ketl.db.settings import get_session
from ketl.db import models


class APIFactory(SQLAlchemyModelFactory):

    class Meta:
        model = models.API
        sqlalchemy_session = get_session()
        sqlalchemy_session_persistence = 'commit'

    name = factory.Sequence(lambda n: f'api_{n}')


class SourceFactory(SQLAlchemyModelFactory):

    class Meta:
        model = models.Source
        sqlalchemy_session = get_session()
        sqlalchemy_session_persistence = 'commit'

    api_config = factory.SubFactory(APIFactory)


class CachedFileFactory(SQLAlchemyModelFactory):

    class Meta:
        model = models.CachedFile
        sqlalchemy_session = get_session()
        sqlalchemy_session_persistence = 'commit'

    source = factory.SubFactory(SourceFactory)


class CredsFactory(SQLAlchemyModelFactory):

    class Meta:
        model = models.Creds
        sqlalchemy_session = get_session()
        sqlalchemy_session_persistence = 'commit'

    api_config = factory.SubFactory(APIFactory)


class ExpectedFileFactory(SQLAlchemyModelFactory):

    class Meta:
        model = models.ExpectedFile
        sqlalchemy_session = get_session()
        sqlalchemy_session_persistence = 'commit'

    cached_file = factory.SubFactory(CachedFileFactory)



#!/usr/bin/env python3
"""Console script for ketl."""
import sys
import click
import alembic.config

from ketl.db import settings


@click.group()
@click.pass_context
def ketl(ctx):

    ctx.ensure_object(dict)

    import os
    print(os.getcwd())


@ketl.command()
@click.option('--db-dsn')
@click.pass_context
def apply_migrations(ctx, db_dsn):

    settings.DB_DSN = db_dsn
    import os
    print(os.getcwd())


if __name__ == "__main__":
    sys.exit(ketl())  # pragma: no cover



import urllib.parse as up
import asyncio

from abc import abstractmethod
from datetime import datetime, timedelta
from dataclasses import dataclass
from ftplib import FTP
from functools import partial
from pathlib import Path
from typing import List, Union, Optional, Iterator
from urllib.parse import quote
from multiprocessing.pool import Pool

from furl import furl
from smart_open import open as smart_open
from tqdm import tqdm
from sqlalchemy.orm import defer, Query
from more_itertools import chunked

from ketl.db.models import API, CachedFile, ExpectedFile, Source
from ketl.db.settings import get_session
from ketl.utils.file_utils import file_hash


class BaseExtractor:

    @abstractmethod
    def extract(self) -> List[Path]:

        raise NotImplementedError('extract not implemented in the base class')


class DefaultExtractor(BaseExtractor):
    """
    The default extractor can fetch files from an FTP server or any location that
    is openable via smart_open. It is up to the user to provide any credentials
    that are required to access the desired resources.
    """
    BLOCK_SIZE = 16384

    def __init__(self, api_config: Union[API, int, str], skip_exiting_files: bool = False,
                 show_progress: bool = False, concurrency: str = 'sync',
                 on_disk_check='full'):

        if type(api_config) is int:
            self.api = get_session().query(API).filter(API.id == api_config).one()
        elif type(api_config) is str:
            self.api = get_session().query(API).filter(API.name == api_config).one()
        elif isinstance(api_config, API):
            self.api = api_config
        self.headers = {}
        self.auth = None
        self.auth_token = None
        self.skip_existing_files = skip_exiting_files
        self.show_progress = show_progress
        self.concurrency = concurrency
        self.on_disk_check = on_disk_check

        if self.api.creds:
            details = self.api.creds.creds_details
            cookie = details.get('cookie', None)
            if cookie:
                self.headers['Cookie'] = cookie['name'] + '=' + cookie['value']
            self.auth = details.get('auth', None)
            self.auth_token = details.get('auth_token', None)
            if self.auth_token:
                self.headers[self.auth_token['header']] = self.auth_token['token']

    def extract(self) -> List[Path]:

        session = get_session()
        
        # depending on whether we are skipping files known to be on disk
        # we produce an iterable that is either a list of queries that will
        # give us the files that are missing, or a chunked version of a query
        if self.skip_existing_files:
            kwargs = {'missing': True, 'use_hash': self.on_disk_check == 'hash'}
            data_iterator: Query = self.api.cached_files_on_disk(**kwargs)
        else:
            data_iterator: Query = self.api.cached_files

        data_iterator = data_iterator.options(defer(CachedFile.meta))
        collected_results = []
        
        for batch in tqdm(chunked(data_iterator, 10000)):  # type: List[CachedFile]

            if self.concurrency == 'sync':
                results = list(
                    filter(None, [
                        self.get_file(
                            cached_file.id, cached_file.full_url, cached_file.full_path,
                            cached_file.refresh_interval, cached_file.url_params,
                            show_progress=self.show_progress
                        ) for cached_file in batch])
                )
                collected_results.extend(results)
                
            elif self.concurrency == 'async':
                raise NotImplementedError('Async downloads not yet implemented.')
            elif self.concurrency == 'multiprocess':
                get_file_args = [(
                    cached_file.id, cached_file.full_url, cached_file.full_path,
                    cached_file.refresh_interval, cached_file.url_params,
                    self.show_progress
                ) for cached_file in batch]

                if get_file_args:
                    with Pool() as pool:
                        futures = pool.starmap_async(self.get_file, get_file_args)
                        results = futures.get()
                        if results:
                            results = list(filter(None, results))
                            collected_results.extend(results)
                    pool.join()
                    
            session.bulk_update_mappings(CachedFile, collected_results)

        session.commit()

        new_expected_files: List[dict] = []
        updated_expected_files: List[dict] = []
        
        q: Query = session.query(
            ExpectedFile.path, ExpectedFile.cached_file_id, ExpectedFile.id
        ).join(
            CachedFile, ExpectedFile.cached_file_id == CachedFile.id
        ).join(
            Source, CachedFile.source_id == Source.id
        ).filter(
            Source.api_config_id == self.api.id
        )

        current_files = {(ef[0], ef[1]): ef[2] for ef in q.yield_per(10000)}

        for source_file in data_iterator:
            ef = source_file.preprocess()
            if ef:
                key = (ef['path'], ef['cached_file_id'])
                if key not in current_files:
                    new_expected_files.append(ef)
                else:
                    updated_expected_files.append({'id': current_files[key], **ef})
        
        session.bulk_insert_mappings(ExpectedFile, new_expected_files)
        session.bulk_update_mappings(ExpectedFile, updated_expected_files)
        session.commit()

        return [Path(ef.path) for ef in self.api.expected_files]

    @classmethod
    def _fetch_ftp_file(cls, source_url: str, target_file: Path, refresh_interval: timedelta,
                        show_progress=False, force_download=False) -> bool:

        parsed_url = up.urlparse(source_url)
        ftp = FTP(parsed_url.hostname)
        ftp.login()
        total_size = ftp.size(parsed_url.path)
        updated = False

        if cls._requires_update(target_file, total_size, refresh_interval) or force_download:

            bar = tqdm(total=total_size, unit='B', unit_scale=True) if show_progress else None

            target_file.parent.mkdir(exist_ok=True, parents=True)

            with open(target_file.as_posix(), 'wb') as f:
                ftp.retrbinary(f'RETR {parsed_url.path}',
                               partial(cls._ftp_writer, f, bar=bar),
                               blocksize=cls.BLOCK_SIZE)
            if bar:
                bar.close()
            updated = True

        return updated

    @classmethod
    def _fetch_generic_file(cls, source_url: str, target_file: Path, refresh_interval: timedelta,
                            url_params=None, headers=None, auth=None,
                            show_progress=False, force_download=False) -> bool:

        transport_params = {}
        url = furl(source_url)
        if headers:
            transport_params['headers'] = headers
        if auth:
            transport_params.update(auth)
        if url_params:
            url.add(url_params)

        # tragic hack that is necessitated by s3's failure to properly conform to http spec
        # c.f. https://forums.aws.amazon.com/thread.jspa?threadID=55746

        url_to_fetch = url.url
        if url.scheme in {'s3', 's3a'} or url.host == 's3.amazonaws.com':
            url_to_fetch = f'{url.scheme}://{url.host}/{quote(str(url.path))}'
            if url.fragmentstr != '':
                url_to_fetch += quote(f'#{url.fragmentstr}', safe='%')
            if url.querystr != '':
                url_to_fetch += f'&{url.querystr}'

        updated = False
        with smart_open(url_to_fetch, 'rb', ignore_ext=True, transport_params=transport_params) as r:
            total_size = getattr(r, 'content_length', -1)
            if cls._requires_update(target_file, total_size, refresh_interval) or force_download:
                target_file.parent.mkdir(exist_ok=True, parents=True)
                bar = tqdm(total=total_size, unit='B', unit_scale=True) if show_progress else None
                with open(target_file.as_posix(), 'wb') as f:
                    # this is actually identical to shutil.copyfileobj(r.raw, f)
                    # but with tqdm injected to show progress
                    cls._generic_writer(r, f, block_size=cls.BLOCK_SIZE, bar=bar)
                updated = True

        return updated

    @staticmethod
    def _ftp_writer(dest, block, bar=None):
        if bar:
            bar.update(len(block))
        dest.write(block)

    @staticmethod
    def _requires_update(target_file: Path, total_size: int, time_delta: timedelta = None) -> bool:

        if target_file.exists():
            stat = target_file.stat()
            existing_size = stat.st_size
            # if either the sizes match up or the size can't be obtained but the file is recent
            return not ((existing_size == total_size)
                        or (total_size == -1 and
                            (datetime.now() - datetime.fromtimestamp(stat.st_mtime)) < time_delta))
        else:
            return True

    @staticmethod
    def _generic_writer(source, target, block_size=16384, bar=None):

        while 1:
            buf = source.read(block_size)
            if bar:
                bar.update(len(buf))
            if not buf:
                break
            target.write(buf)
        if bar:
            bar.close()

    def get_file(self, cached_file_id: int, source_url: str, target_file: Path, refresh_interval: timedelta,
                 url_params=None, show_progress=False, force_download=False) -> Optional[dict]:

        try:
            parsed_url = up.urlparse(source_url)
            if parsed_url.scheme == 'ftp':
                result = self._fetch_ftp_file(source_url,
                                              target_file,
                                              refresh_interval,
                                              show_progress=show_progress,
                                              force_download=force_download)
            else:
                result = self._fetch_generic_file(source_url,
                                                  target_file,
                                                  refresh_interval,
                                                  url_params,
                                                  headers=self.headers,
                                                  auth=self.auth,
                                                  show_progress=show_progress,
                                                  force_download=force_download)

            if result:
                return {
                    'id': cached_file_id,
                    'hash': file_hash(target_file).hexdigest(),
                    'last_download': datetime.now(),
                    'size': target_file.stat().st_size
                }
            else:
                return None

        except Exception as ex:
            print(f'Could not download {source_url}: {ex}')
            return None

    @staticmethod
    def _update_file_cache(source_file: CachedFile, target_file: Path):

        session = get_session()
        source_file.hash = file_hash(target_file).hexdigest()
        source_file.last_download = datetime.now()
        source_file.size = target_file.stat().st_size
        session.add(source_file)
        session.commit()



import pytest

from marshmallow import Schema, fields, ValidationError
from unittest import mock

from ketl.extractor.Rest import RestMixin


class JsonParams(Schema):

    param1 = fields.Int()
    param2 = fields.Str()


class JsonResponse(Schema):

    foo = fields.Str()
    bar = fields.Int()


@pytest.fixture
def rest():

    return RestMixin()


@pytest.fixture
def good_json_response():

    return {'foo': 'hello', 'bar': 1}


@pytest.fixture
def bad_json_response():

    return {'foo': 1, 'bar': 'hello'}


@mock.patch('ketl.extractor.Rest.requests')
def test_rest_get(mock_requests, good_json_response, bad_json_response, rest):

    mock_json_response = mock.Mock()
    mock_json_response.json.return_value = good_json_response
    mock_requests.get.return_value = mock_json_response

    result = rest.get('http://foo', 'bar', params={'param1': 1, 'param2': 'foo'}, data_schema=JsonParams(),
                      result_schema=JsonResponse())

    assert result == good_json_response

    result = rest.get('http://foo', 'bar', params={'param1': 1, 'param2': 'foo'}, data_schema=JsonParams())

    assert result == good_json_response

    mock_json_response.json.return_value = bad_json_response

    with pytest.raises(ValidationError):
        result = rest.get('http://foo', 'bar', params={'param1': 1, 'param2': 'foo'}, data_schema=JsonParams(),
                          result_schema=JsonResponse())


@mock.patch('ketl.extractor.Rest.requests')
def test_rest_post(mock_requests, good_json_response, bad_json_response, rest):

    mock_json_response = mock.Mock()
    mock_json_response.json.return_value = good_json_response
    mock_requests.post.return_value = mock_json_response

    result = rest.post('http://foo', 'bar', json={'param1': 1, 'param2': 'foo'}, data_schema=JsonParams(),
                       result_schema=JsonResponse())

    assert result == good_json_response

    result = rest.post('http://foo', 'bar', json={'param1': 1, 'param2': 'foo'}, data_schema=JsonParams())

    assert result == good_json_response

    mock_json_response.json.return_value = bad_json_response

    with pytest.raises(ValidationError):
        result = rest.post('http://foo', 'bar', json={'param1': 1, 'param2': 'foo'}, data_schema=JsonParams(),
                           result_schema=JsonResponse())


@mock.patch('ketl.extractor.Rest.requests')
def test_rest_put(mock_requests, good_json_response, bad_json_response, rest):

    mock_json_response = mock.Mock()
    mock_json_response.json.return_value = good_json_response
    mock_requests.put.return_value = mock_json_response

    result = rest.put('http://foo', 'bar', json={'param1': 1, 'param2': 'foo'}, data_schema=JsonParams(),
                      result_schema=JsonResponse())

    assert result == good_json_response

    result = rest.put('http://foo', 'bar', json={'param1': 1, 'param2': 'foo'}, data_schema=JsonParams())

    assert result == good_json_response

    mock_json_response.json.return_value = bad_json_response

    with pytest.raises(ValidationError):
        result = rest.put('http://foo', 'bar', json={'param1': 1, 'param2': 'foo'}, data_schema=JsonParams(),
                          result_schema=JsonResponse())


def test_unknown_method(rest):

    with pytest.raises(ValueError):
        rest._execute_request('http://foo', 'delete')



from abc import abstractmethod
from enum import Enum
from hashlib import sha256
from pathlib import Path
from typing import Union, Any, Callable

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import pickle

from sqlalchemy import text

from ketl.db.settings import get_engine


class InvalidLoaderConfiguration(Exception):
    pass


class BaseLoader:

    def __init__(self, destination: Union[Path, str], **kwargs):

        self.destination = destination

    @abstractmethod
    def load(self, data_frame: pd.DataFrame):

        raise NotImplementedError

    @abstractmethod
    def finalize(self):

        raise NotImplementedError


class HashLoader(BaseLoader):

    def load(self, data_frame: pd.DataFrame, **kwargs):

        df_hash = sha256(pd.util.hash_pandas_object(data_frame).values).hexdigest()
        with open(self.destination, 'w') as f:
            f.write(df_hash + '\n')

    def finalize(self):
        pass


class LocalFileLoader(BaseLoader):

    def __init__(self, destination: Union[Path, str], naming_func: Callable = None, **kwargs):

        super().__init__(Path(destination))
        self.naming_func = naming_func
        self.writer = None
        self.kwargs = kwargs

        # deletes compatible with pre-3.8 python

        if self.destination.is_dir():
            files = self.destination.glob('*')
            for file in files:
                if file.exists():
                    file.unlink()
        else:
            if self.destination.exists():
                self.destination.unlink()

    def full_path(self, df: pd.DataFrame):
        if not self.naming_func:
            return self.destination
        else:
            return self.destination / self.naming_func(df)

    def finalize(self):

        pass


class ParquetLoader(LocalFileLoader):

    def load(self, data_frame: pd.DataFrame):

        try:
            table = pa.Table.from_pandas(data_frame)
            if not self.writer:
                self.writer = pq.ParquetWriter(self.full_path(data_frame), table.schema)
            else:
                if not self.full_path(data_frame).exists():
                    self.writer.close()
                    self.writer = pq.ParquetWriter(self.full_path(data_frame), table.schema)
            self.writer.write_table(table)
        except Exception as ex:
            print(f'Could not process {self.full_path(data_frame)}')
            raise ex

    def finalize(self):

        if self.writer:
            self.writer.close()


class DelimitedFileLoader(LocalFileLoader):

    def load(self, data_frame: pd.DataFrame):
        with open(self.full_path(data_frame), 'a') as f:
            data_frame.to_csv(f, **self.kwargs)


class DatabaseLoader(BaseLoader):

    def __init__(self, destination: str, **kwargs):

        self.engine = get_engine()
        self.schema = kwargs.pop('schema', None)
        self.kwargs = kwargs
        super().__init__(destination)

        if self.schema:
            self.delete_stmt = text(f'DELETE FROM {self.schema}.{self.destination}')
        else:
            self.delete_stmt = text(f'DELETE FROM {self.destination}')

        self.clean = False

    def load(self, data_frame: pd.DataFrame):

        # this is somewhat less general than either having a proper pre/post load hook
        # logic or doing something like writing to a temp staging table and then moving
        # the final result to the desired destination, but it's also *a lot* simpler

        if not self.clean:
            self.engine.execute(self.delete_stmt)
            self.clean = True

        data_frame.to_sql(self.destination, self.engine, index=False, if_exists='append', schema=self.schema)

    def finalize(self):

        pass


class PickleLoader(BaseLoader):

    def __init__(self, pickler=None, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.pickler = pickler or pickle.dump

    def load(self, obj: Any):

        with open(self.destination, 'wb') as f:
            self.pickler(obj, f)

    def finalize(self):

        pass



import pytest
import pandas as pd

from hashlib import sha256
from tempfile import NamedTemporaryFile

from ketl.loader.Loader import BaseLoader, DatabaseLoader, DataFrameLoader, HashLoader
from ketl.db.settings import get_engine


@pytest.fixture
def data_frame():

    return pd.DataFrame.from_records([(1, 2, 3), (4, 5, 6)], columns=['x', 'y', 'z'])


def test_base_loader(data_frame):

    loader = BaseLoader('foo')

    assert loader.destination == 'foo'

    with pytest.raises(NotImplementedError):
        loader.load(data_frame)

    with pytest.raises(NotImplementedError):
        loader.finalize()


def test_hash_loader(data_frame, tmp_path):

    tf = NamedTemporaryFile(dir=tmp_path, delete=False)
    tf.close()

    loader = HashLoader(tf.name)

    expected_hash = sha256(pd.util.hash_pandas_object(data_frame).values).hexdigest()

    loader.load(data_frame)

    with open(tf.name, 'r') as f:
        data = f.read().strip()
        assert expected_hash == data

    loader.finalize()


def test_data_frame_loader_csv(data_frame, tmp_path):

    csv_file = tmp_path / 'df.csv'
    with open(csv_file, 'w') as f:
        f.write('hello world')

    loader = DataFrameLoader(csv_file, index=False)

    assert loader.dest_path == csv_file
    assert loader.file_format == DataFrameLoader.FileFormat.CSV
    assert loader.kwargs == {'index': False}
    assert not   csv_file.exists()

    loader.load(data_frame)

    assert csv_file.exists()

    df_out = pd.read_csv(csv_file, index_col=None)

    assert df_out.equals(data_frame)


def test_data_frame_loader_parquet(data_frame, tmp_path):

    parquet_file = tmp_path / 'df.parquet'

    loader = DataFrameLoader(parquet_file, index=False)

    assert loader.dest_path == parquet_file
    assert loader.file_format == DataFrameLoader.FileFormat.PARQUET
    assert loader.kwargs == {'index': False}
    assert not parquet_file.exists()

    loader.load(data_frame)
    loader.finalize()

    assert parquet_file.exists()

    df_out = pd.read_parquet(parquet_file)

    assert df_out.equals(data_frame)


def test_data_frame_loader_unknown():

    with pytest.raises(ValueError):
        loader = DataFrameLoader('some-file.unknown')


def test_database_loader_init():

    loader = DatabaseLoader('test_table', index=False)

    assert loader.schema is None
    assert loader.kwargs == {'index': False}
    assert not loader.clean

    loader = DatabaseLoader('test_table', schema='test_schema')

    assert loader.schema == 'test_schema'
    assert loader.kwargs == {}
    assert not loader.clean


def test_database_loader(data_frame):

    engine = get_engine()
    with engine.connect() as conn:
        conn.execute('DROP TABLE IF EXISTS test_table')
        conn.execute('CREATE TABLE IF NOT EXISTS test_table (x INTEGER, y INTEGER, z INTEGER)')

    loader = DatabaseLoader('test_table', index=False)

    assert loader.engine == engine
    assert loader.schema is None
    assert loader.kwargs == {'index': False}

    loader.load(data_frame)

    df_out = pd.read_sql_table('test_table', engine.connect(), index_col=False)

    assert df_out.equals(data_frame)



import pytest

from unittest import mock
from pathlib import Path

from ketl.db.settings import get_session
from ketl.etl.Pipeline import ETLPipeline, InvalidPipelineError
from ketl.extractor.Extractor import DefaultExtractor
from ketl.transformer.Transformer import DelimitedTableTransformer
from ketl.loader.Loader import DataFrameLoader
from tests.factories import APIFactory


def test_pipeline_init_without_fanout():

    api = APIFactory(name='my nice api')

    ex1 = DefaultExtractor(api)
    ex2 = DefaultExtractor(api)
    tf1 = DelimitedTableTransformer()
    tf2 = DelimitedTableTransformer()
    l1 = DataFrameLoader('out.csv')
    l2 = DataFrameLoader('out.csv')

    pipeline = ETLPipeline(extractors=[ex1, ex2], transformers=[tf1, tf2], loaders=[l1, l2])

    for k, v in pipeline.fanout.items():
        if k == ex1:
            assert pipeline.fanout[k] == [tf1, tf2]
        if k == ex2:
            assert pipeline.fanout[k] == [tf1, tf2]
        if k == tf1:
            assert pipeline.fanout[k] == [l1, l2]
        if k == tf2:
            assert pipeline.fanout[k] == [l1, l2]


def test_pipeline_init_with_fanout(tmp_path):

    api = APIFactory(name='my nice api')

    ex1 = DefaultExtractor(api)
    ex2 = DefaultExtractor(api)
    tf1 = DelimitedTableTransformer()
    tf2 = DelimitedTableTransformer()
    l1 = DataFrameLoader('out.csv')
    l2 = DataFrameLoader('out.csv')

    fanout = {
        ex1: [tf1, tf2],
        tf1: [l1],
        tf2: [l2]
    }

    pipeline = ETLPipeline(extractors=[ex1, ex2], transformers=[tf1, tf2], loaders=[l1, l2], fanout=fanout)

    assert pipeline.fanout == fanout


def test_pipeline_init_bad_fanout():

    api = APIFactory(name='my nice api')

    ex1 = DefaultExtractor(api)
    ex2 = DefaultExtractor(api)
    tf1 = DelimitedTableTransformer()
    tf2 = DelimitedTableTransformer()
    l1 = DataFrameLoader('out.csv')
    l2 = DataFrameLoader('out.csv')

    fanout = {
        ex1: [tf1, l1],
        tf1: [l1, l2],
        tf2: [l2]
    }

    with pytest.raises(InvalidPipelineError):
        pipeline = ETLPipeline(fanout=fanout)

    fanout = {
        ex1: [tf1, tf2],
        tf1: [l1, l2],
        tf2: [tf2]
    }

    with pytest.raises(InvalidPipelineError):
        pipeline = ETLPipeline(fanout=fanout)


@mock.patch('ketl.extractor.Extractor.DefaultExtractor.extract')
def test_fire_extractors(mock_extract):

    api = APIFactory(name='my nice api')

    ex1 = DefaultExtractor(api)
    ex2 = DefaultExtractor(api)
    tf1 = DelimitedTableTransformer()
    tf2 = DelimitedTableTransformer()

    pipeline = ETLPipeline(extractors=[ex1, ex2], transformers=[tf1, tf2])

    mock_extract.side_effect = [[Path('file1'), Path('file2')], [Path('file3')]]

    result = pipeline._fire_extractors()

    assert result == {ex1: [Path('file1'), Path('file2')], ex2: [Path('file3')]}


@mock.patch('ketl.loader.Loader.DataFrameLoader.load')
@mock.patch('ketl.transformer.Transformer.DelimitedTableTransformer.transform')
def test_fire_transformers(mock_transform: mock.Mock, mock_load: mock.Mock):

    api = APIFactory(name='my nice api')

    ex1 = DefaultExtractor(api)
    tf1 = DelimitedTableTransformer()
    l1 = DataFrameLoader('out.csv')

    pipeline = ETLPipeline(extractors=[ex1], transformers=[tf1], loaders=[l1])

    extraction_results = {ex1: [Path('file1')]}

    mock_transform.return_value = [mock.Mock()]
    mock_load.return_value = None

    pipeline._fire_transformers(extraction_results)

    mock_transform.assert_called_once()
    mock_load.assert_called_once()



import pandas as pd

from typing import Dict, List, Callable, Optional, Union, Type
from itertools import product
from pathlib import Path
from collections import defaultdict

from ketl.extractor.Extractor import BaseExtractor
from ketl.transformer.Transformer import BaseTransformer
from ketl.loader.Loader import BaseLoader


class InvalidPipelineError(Exception):
    pass


class ETLPipeline:

    def __init__(self, extractors: List[BaseExtractor] = None,
                 transformers: List[BaseTransformer] = None,
                 loaders: List[BaseLoader] = None,
                 fanout: Dict[Union[BaseExtractor, BaseTransformer, BaseLoader],
                              Union[List[BaseTransformer], List[BaseLoader]]] = None):

        extractors = extractors or []
        transformers = transformers or []
        loaders = loaders or []

        if not fanout:
            self.fanout: Dict[Union[BaseExtractor, BaseTransformer, BaseLoader],
                              Union[List[BaseTransformer], List[BaseLoader]]] = defaultdict(list)
            for extractor, transformer in product(extractors, transformers):
                self.fanout[extractor].append(transformer)
            for transformer, loader in product(transformers, loaders):
                self.fanout[transformer].append(loader)
        else:
            for k, v in fanout.items():
                if isinstance(k, BaseExtractor) and not all(map(lambda item: isinstance(item, BaseTransformer), v)):
                    raise InvalidPipelineError(f'An Extractor {k} is attached to something other than a Transformer: {v}.')
                elif isinstance(k, BaseTransformer) and not all(map(lambda item: isinstance(item, BaseLoader), v)):
                    raise InvalidPipelineError(f'A Transformer {k} is attached to something other than a Loader: {v}.')
            self.fanout = fanout

    def execute(self):

        extraction_results = self._fire_extractors()
        self._fire_transformers(extraction_results)

    def _fire_extractors(self) -> Dict[BaseExtractor, List[Path]]:

        results = {op: op.extract() for op in self.fanout if isinstance(op, BaseExtractor)}

        return results

    def _fire_transformers(self, extraction_results: Dict[BaseExtractor, List[Path]]):

        for op in self.fanout:
            if isinstance(op, BaseExtractor):
                result = extraction_results[op]
                transformers = self.fanout.get(op, [])
                # TODO: parallelize this using joblib or something like that
                for transformer in transformers:
                    loaders = self.fanout.get(transformer, [])
                    for df in transformer.transform(result):  # type: pd.DataFrame
                        if not df.empty:
                            for loader in loaders:  # type: BaseLoader
                                loader.load(df)
                    for loader in loaders:
                        loader.finalize()



import requests
import enum
import tarfile
import zipfile
import gzip
import lzma
import shutil

from abc import abstractmethod
from datetime import timedelta
from hashlib import sha1
from pathlib import Path
from furl import furl
from typing import Optional, Set, List, Dict, Type
from marshmallow import Schema
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import (
    Column, Boolean, Integer, String, ForeignKey, DateTime,
    JSON, Enum, Interval, UniqueConstraint, BigInteger, Index
)
from more_itertools import chunked
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import relationship, joinedload, Query

from ketl.extractor.Rest import RestMixin
from ketl.utils.file_utils import file_hash
from ketl.utils.db_utils import get_or_create
from ketl.db.settings import get_engine, get_session


Base = declarative_base()


class API(Base, RestMixin):
    """
    The API class is the center of the organizational model for kETL. It configures the basic logic
    of accessing some set of resources, setting up credentials as needed.
    """

    __tablename__ = 'ketl_api_config'

    id = Column(Integer, primary_key=True)
    name = Column(String, index=True, unique=True)
    description = Column(String, index=True)
    sources = relationship('Source', back_populates='api_config', enable_typechecks=False)
    creds = relationship('Creds', back_populates='api_config', lazy='joined', uselist=False)
    hash = Column(String)

    BATCH_SIZE = 10000

    @abstractmethod
    def setup(self):
        """
        All subclasses of API must implement the setup method to generate the actual configuration
        that will specify what is to be downloaded.
        :return:
        """
        if not self.name:
            self.name = self.__class__.__name__

        session = get_session()
        existing_api = session.query(API).filter(API.name == self.name).one_or_none()
        if not existing_api:
            session.add(self)
            session.commit()
        else:
            self.id = existing_api.id

    @property
    def api_hash(self) -> str:
        """
        Hash the API by hashing all of its sources and return the hex digest.
        :return: Hex digest of the hash.
        """

        s = sha1(bytes(self.name, 'utf-8'))
        for source in self.sources:
            s.update(source.source_hash.digest())

        return s.hexdigest()

    @staticmethod
    def get_instance(model: Type['API'], name=None) -> 'API':
        """
        Retrieve an instance of the given subclass of API. There can only be one instance per name.
        :param model: A subclass of API.
        :param name: An optional name. Only one API per name is allowed.
        :return: An instance of the provided subclass of API.
        """
        name = name or model.__name__
        instance, created = get_or_create(model, name=name)
        return instance

    @property
    def expected_files(self) -> Query:
        """
        Retrieve all the expected files under this API.
        :return: A list of expected files.
        """

        q = get_session().query(
            ExpectedFile
        ).join(
            CachedFile, ExpectedFile.cached_file_id == CachedFile.id
        ).join(
            Source, CachedFile.source_id == Source.id
        ).filter(
            Source.api_config_id == self.id
        )

        return q.yield_per(self.BATCH_SIZE)

    @property
    def cached_files(self) -> Query:

        q: Query = get_session().query(
            CachedFile
        ).join(
            Source, CachedFile.source_id == Source.id
        ).filter(
            Source.api_config_id == self.id
        )

        return q.yield_per(self.BATCH_SIZE)

    def cached_files_on_disk(self, use_hash=True, missing=False, limit_ids=None) -> Query:

        # actually checking existence can get pretty expensive for the
        # scenario where you have lots of files on a distributed file system
        # so if we're pretty sure we haven't deleted the files manually, we
        # can trust that if a hash is present in the db, then the file exists

        files = get_session().query(
            CachedFile
        ).join(
            Source, CachedFile.source_id == Source.id
        ).filter(
            Source.api_config_id == self.id
        )

        if use_hash:
            if missing:
                files = files.filter(CachedFile.hash.is_(None))
            else:
                files = files.filter(CachedFile.hash.isnot(None))

            if limit_ids:
                files = files.filter(CachedFile.id.in_(limit_ids))

            return files.yield_per(self.BATCH_SIZE)

        else:
            q: Query = get_session().query(
                Source.data_dir, CachedFile.path, CachedFile.id
            ).join(
                Source, CachedFile.source_id == Source.id
            ).filter(
                Source.api_config_id == self.id
            )

            file_ids = set()
            for batch in chunked(q.yield_per(self.BATCH_SIZE), self.BATCH_SIZE):
                file_ids = {item[2] for item in batch if (Path(item[0]) / item[1]).resolve().exists()}

            if limit_ids:
                file_ids = file_ids & limit_ids

            if missing:
                file_ids = {item[2] for item in batch} - file_ids

            return files.filter(CachedFile.id.in_(file_ids)).yield_per(self.BATCH_SIZE)


class ExpectedMode(enum.Enum):

    auto = 'auto'
    explicit = 'explicit'
    self = 'self'


class InvalidConfigurationError(Exception):
    """Exception indicating an invalid configuration."""


class CachedFile(Base):
    """
    The CachedFile class represents a single file that may be downloaded by an extractor.
    """

    BLOCK_SIZE = 65536

    __tablename__ = 'ketl_cached_file'

    __table_args__ = (
        UniqueConstraint('source_id', 'url', 'path'),
        Index('ix_ketl_cached_file_meta', 'meta', postgresql_using='gin')
    )

    id = Column(Integer, primary_key=True)
    source_id = Column(Integer, ForeignKey('ketl_source.id', ondelete='CASCADE'))
    source = relationship('Source', back_populates='source_files')
    expected_files = relationship('ExpectedFile', back_populates='cached_file')
    url = Column(String, index=True)
    url_params = Column(JSONB)
    path = Column(String, index=True)  # path relative to source
    last_download = Column(DateTime, nullable=True, index=True)
    last_update = Column(DateTime, nullable=True, index=True)
    refresh_interval = Column(Interval, nullable=True, index=True, default=timedelta(days=7))
    hash = Column(String, nullable=True)
    cache_type = Column(String, index=True, nullable=True)
    size = Column(BigInteger, index=True, nullable=True)
    is_archive = Column(Boolean, index=True, default=False)
    extract_to = Column(String, index=True, nullable=True)
    expected_mode = Column(Enum(ExpectedMode), index=True, default=ExpectedMode.explicit)
    meta = Column(JSONB, nullable=True)

    @property
    def full_path(self) -> Path:
        """
        Return the absolute path of the cached file.
        :return: The absolute path of the file.
        """
        return Path(self.source.data_dir).resolve() / self.path

    @property
    def full_url(self) -> str:

        return f'{self.source.base_url}/{self.url}'

    @property
    def file_hash(self):
        """
        Return the hash of the file.
        :return: The hash object (not the digest or the hex digest!) of the file.
        """
        if self.path:
            path = Path(self.path).resolve()
            if path.exists() and not path.is_dir():
                return file_hash(path, self.BLOCK_SIZE)

        return sha1()

    def preprocess(self) -> Optional[dict]:
        """
        Preprocess the file, extracting and creating expected files as needed.
        :return: Optionally returns an expected file, if one was created directly from the
            cached file. Otherwise returns None.
        """
        extract_dir = Path(self.extract_to) if self.extract_to is not None and self.extract_to != '' else Path('.')

        if self.is_archive:
            expected_paths: Set[Path] = {Path(file.path) for file in self.expected_files}
            if tarfile.is_tarfile(self.full_path):
                self._extract_tar(extract_dir, expected_paths)
            elif zipfile.is_zipfile(self.full_path):
                self._extract_zip(extract_dir, expected_paths)
            elif self.full_path.name.endswith('.gz'):
                self._extract_gzip(extract_dir)
            elif self.full_path.suffix in ['.xz', '.lz', '.lzma']:
                self._extract_lzma(extract_dir)
            return None
        elif self.expected_mode == ExpectedMode.self:
            return {'cached_file_id': self.id, 'path': str(self.full_path)}

    def _extract_tar(self, extract_dir: Path, expected_paths: Set[Path]):
        """
        Extracts a tarball into the target directory. Creates expected files as needed.
        :param extract_dir: The directory to which the tarball is to be extracted.
        :param expected_paths: The list of expected paths that should be generated from the archive.
        :return: None
        """
        tf = tarfile.open(self.full_path)
        if self.expected_mode == ExpectedMode.auto:
            archived_paths = {Path(file) for file in tf.getnames()}
            self._generate_expected_files(extract_dir, archived_paths, expected_paths)
            tf.extractall(path=extract_dir)
        elif self.expected_mode == ExpectedMode.explicit:
            for file in self.expected_files:
                target = extract_dir / file.path
                with open(target, 'wb') as target_file:
                    with tf.extractfile(file.archive_path) as source_file:
                        shutil.copyfileobj(source_file, target_file)

    def _extract_zip(self, extract_dir: Path, expected_paths: Set[Path]):
        """
        Extracts a zip archive into the target directory. Creates expected files as needed.
        :param extract_dir: The directory to which the archive is to be extracted.
        :param expected_paths: The list of expected paths that should be generated from the archive.
        :return: None
        """
        zf = zipfile.ZipFile(self.full_path)
        archived_paths = {Path(file) for file in zf.namelist()}
        if self.expected_mode == ExpectedMode.auto:
            self._generate_expected_files(extract_dir, archived_paths, expected_paths)
            zf.extractall(path=extract_dir)
        elif self.expected_mode == ExpectedMode.explicit:
            for file in self.expected_files:
                target = extract_dir / file.path
                info = zf.getinfo(file.archive_path)
                if not target.exists() or info.file_size != target.stat().st_size:
                    zf.extract(file.archive_path, path=extract_dir)
                    source = (extract_dir / file.archive_path).resolve()
                    if source != target.resolve():
                        shutil.move(source, target)

    def _determine_target(self, extract_dir: Path) -> Path:

        if len(self.expected_files) > 1:
            raise InvalidConfigurationError(f'More than 1 expected file configured for a gz archive: {self.path}')
        if len(self.expected_files) == 0 and self.expected_mode == ExpectedMode.auto:
            return extract_dir / Path(self.path).stem
        elif len(self.expected_files) == 0 and not self.expected_mode == ExpectedMode.auto:
            raise InvalidConfigurationError(f'Expected mode is set to {self.expected_mode} but '
                                            f'no expected files supplied.')
        elif len(self.expected_files) == 1 and self.expected_mode == ExpectedMode.explicit:
            return extract_dir / self.expected_files[0].path
        else:
            raise InvalidConfigurationError('Something very bad has happened :(')

    def _extract_gzip(self, extract_dir: Path) -> None:
        """
        Extracts a gz file into the target directory.
        :param extract_dir: The directory to which the archive is to be extracted.
        :return: None
        """
        result_file = self._determine_target(extract_dir)

        with open(result_file, 'wb') as target:
            with gzip.open(self.full_path, 'r') as source:
                shutil.copyfileobj(source, target)

    def _extract_lzma(self, extract_dir: Path) -> None:
        """
        Extracts an lzma file into the target directory.
        :param extract_dir: The directory to which the archive is to be extracted.
        :return: None
        """
        result_file = self._determine_target(extract_dir)

        with open(result_file, 'wb') as target:
            with lzma.open(self.full_path, 'r') as source:
                shutil.copyfileobj(source, target)

    def _generate_expected_files(self, extract_dir: Path, archived_paths: Set[Path], expected_paths: Set[Path]) -> None:
        """
        Generates expected file entries in the table if they do not already exist.
        :param extract_dir: The directory to which the archive is to be extracted.
        :param archived_paths: The list of paths contained in the archive.
        :param expected_paths: The list of expected files.
        :return: None
        """
        session = get_session()

        missing_paths = {path for path in archived_paths if extract_dir / path not in expected_paths}
        expected_files = [ExpectedFile(path=str(extract_dir / path), cached_file=self)
                          for path in missing_paths]
        session.bulk_save_objects(expected_files)


class Creds(Base):
    """
    A simple class for keeping track of credentials. Details are stored in a JSONB blob.

    SECURITY WARNING: creds are currently stored unencrypted. Don't put anything in here
    that requires real security.
    """

    __tablename__ = 'ketl_creds'

    id = Column(Integer, primary_key=True)
    api_config_id = Column(Integer, ForeignKey('ketl_api_config.id', ondelete='CASCADE'))
    api_config = relationship('API', back_populates='creds', enable_typechecks=False)
    creds_details = Column(JSONB)


class Source(Base):
    """
    A class representing a source of some data. Can be subclassed on source type.
    """

    __tablename__ = 'ketl_source'

    __table_args__ = (
        UniqueConstraint('base_url', 'data_dir', 'api_config_id'),
        Index('ix_ketl_source_meta', 'meta', postgresql_using='gin')
    )

    id = Column(Integer, primary_key=True)
    source_type = Column(String, index=True)
    base_url = Column(String, index=True)
    data_dir = Column(String, index=True)
    api_config_id = Column(Integer, ForeignKey('ketl_api_config.id', ondelete='CASCADE'))
    api_config = relationship('API', back_populates='sources', enable_typechecks=False)
    meta = Column(JSONB, nullable=True)
    source_files = relationship('CachedFile', back_populates='source',
                                cascade='all, delete-orphan',
                                passive_deletes=True,
                                lazy='joined')

    __mapper_args__ = {
        'polymorphic_identity': 'source',
        'polymorphic_on': source_type
    }

    @property
    def expected_files(self):
        return [expected_file for cached_file in self.source_files
                for expected_file in cached_file.expected_files]

    @property
    def source_hash(self):

        s = sha1(bytes(self.base_url + self.data_dir, 'utf-8'))
        for source_file in self.source_files:
            s.update(source_file.file_hash.digest())

        return s


class ExpectedFile(Base):
    """
    A class representing expected files to actually be processed.
    """
    __tablename__ = 'ketl_expected_file'

    __table_args__ = (
        UniqueConstraint('path', 'cached_file_id'),
        Index('ix_ketl_expected_file_meta', 'meta', postgresql_using='gin')
    )

    BLOCK_SIZE = 65536

    id = Column(Integer, primary_key=True)
    path = Column(String, index=True)
    archive_path = Column(String, index=True)
    hash = Column(String)
    size = Column(BigInteger, index=True)
    cached_file_id = Column(Integer, ForeignKey('ketl_cached_file.id', ondelete='CASCADE'))
    cached_file = relationship('CachedFile', back_populates='expected_files', lazy='joined')
    processed = Column(Boolean, default=False, index=True)
    file_type = Column(String, index=True)
    last_processed = Column(DateTime, index=True)
    meta = Column(JSONB, nullable=True)

    @property
    def file_hash(self):
        """
        Hash the expected file.
        :return: The hash object.
        """
        if self.path:
            path = Path(self.path).resolve()
            if path.exists() and not path.is_dir():
                return file_hash(path, self.BLOCK_SIZE)

        return sha1()




from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.exc import IntegrityError
from sqlalchemy import MetaData

from ketl.db.settings import get_session


def get_or_create(model, session=None, **kwargs):

    session = session or get_session()
    try:
        return session.query(model).filter_by(**kwargs).one(), False
    except NoResultFound:
        created = model(**kwargs)
        try:
            session.add(created)
            session.commit()
            return created, True
        except IntegrityError:
            session.rollback()
            return session.query(model).filter_by(**kwargs).one(), False


def merge_meta(*metas) -> MetaData:
    merged = MetaData()
    for meta in metas:
        for table in meta.tables.values():
            table.tometadata(merged)
    return merged



import os
import configparser

from pathlib import Path
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import Session


CONFIG_FILE = Path('ketl.ini')

if CONFIG_FILE.exists():
    parser = configparser.ConfigParser()
    parser.read(CONFIG_FILE)
    if 'ketl' in parser:
        DB_DSN = parser.get('ketl', 'DB_DSN', fallback=os.getenv('DB_DSN', 'sqlite:///ketl.db'))
    else:
        DB_DSN = os.getenv('DB_DSN', 'sqlite:///ketl.db')
else:
    DB_DSN = os.getenv('DB_DSN', 'sqlite:///ketl.db')


def get_engine(conn_string=DB_DSN):
    # extremely primitive memoization to get a single global engine object
    return get_session.__dict__.setdefault(f'_engine:{conn_string}', create_engine(conn_string))


def get_session() -> Session:
    # extremely primitive memoization to get a single global session object
    s_maker = get_session.__dict__.setdefault('_sessionmaker', sessionmaker(bind=get_engine()))
    return get_session.__dict__.setdefault('_session', s_maker())




from abc import abstractmethod
from itertools import chain
from pathlib import Path
from typing import List, Tuple, Union, Optional, Dict, Callable
from tqdm import tqdm

import pandas as pd
import numpy as np
import json
import io
import inflection

from ketl.db.settings import get_session


class AdapterError(Exception):
    pass


class NoValidSourcesError(AdapterError):
    pass


class BaseTransformer:

    # TODO: init should take the configuration kwargs

    def __init__(self, transpose: bool = False, concat_on_axis: Union[int, str] = None,
                 columns: List[Union[str, int]] = None, skip_errors: bool = False,
                 rename: Union[Callable, Dict[str, str]] = None, **kwargs):

        self.transpose = transpose
        self.concat_on_axis = concat_on_axis
        self.columns = columns
        self.skip_errors = skip_errors
        self.rename = rename

        self.passed_kwargs = kwargs

    @abstractmethod
    def transform(self, source_files: List[Path]) -> pd.DataFrame:

        raise NotImplementedError

    @abstractmethod
    def _build_data_frame(self, source_files: List[Path]) -> pd.DataFrame:

        raise NotImplementedError


class DelimitedTableTransformer(BaseTransformer):

    def __init__(self, transpose: bool = False, concat_on_axis: Union[str, int] = None,
                 columns: List[Union[str, int]] = None, skip_errors: bool = False,
                 rename: Union[Callable, Dict[str, str]] = None, **kwargs):

        super(DelimitedTableTransformer, self).__init__(
            transpose, concat_on_axis, columns, skip_errors, rename, **kwargs)

        self.reader_kwargs = {
            'comment': None,
            'names': None,
            'delimiter': None,
            'header': 'infer',
            'dtype': None,
            'index_col': None,
            'parse_dates': None,
            'skiprows': None,
            'iterator': True,
            'chunksize': 50000
        }
        self.reader_kwargs.update(self.passed_kwargs)

    def _build_data_frame(self, source_files: List[Path]):

        data_frames = [pd.read_csv(source_file, **self.reader_kwargs) for source_file in source_files]

        # for the special case where every file is a column. this assumes all data can fit into memory
        # TODO: replace this with dask stuff so that things can be lazily concatenated
        if self.concat_on_axis:
            df = pd.concat(data_frames, axis=self.concat_on_axis)
            yield df
        else:
            df_chain = chain(*data_frames)

            for chunk in df_chain:
                if self.transpose:
                    yield chunk.transpose()
                else:
                    yield chunk

    def transform(self, source_files: List[Path]) -> pd.DataFrame:

        for df in self._build_data_frame(source_files):
            yield df


class JsonTableTransformer(BaseTransformer):

    def __init__(self, record_path: Union[List[str], str] = None,
                 transpose: bool = False,
                 concat_on_axis: Union[str, int] = None,
                 columns: List[Union[str, int]] = None,
                 skip_errors: bool = False,
                 rename: Union[Callable, Dict[str, str]] = None,
                 **kwargs):

        super(JsonTableTransformer, self).__init__(
            transpose, concat_on_axis, columns, skip_errors, rename, **kwargs)

        self.record_path = record_path

        self.reader_kwargs = {
            'orient': None,
            'typ': 'frame',
            'dtype': None,
            'convert_axes': None,
            'convert_dates': True,
            'keep_default_dates': True,
            'precise_float': False,
            'date_unit': None,
            'encoding': None,
            'lines': False,
            'chunksize': None,
            'compression': 'infer',
            'nrows': None,
            'storage_options': None
        }
        self.reader_kwargs.update(self.passed_kwargs)

    @staticmethod
    def _extract_data(filename: Union[Path, str], record_path: Union[List[str], str],
                      serialize: bool = True) -> Union[dict, list, str]:

        with open(filename, 'r') as f:
            data: dict = json.load(f)
            if type(record_path) is str:
                if serialize:
                    return json.dumps(data[record_path])
                else:
                    return data[record_path]
            elif type(record_path) is list:
                for item in record_path:
                    data = data[item]
                if serialize:
                    return json.dumps(data)
                else:
                    return data
            else:
                raise TypeError('record_path must be a list or a string')

    def _build_data_frame(self, source_files: List[Path]) -> pd.DataFrame:

        # we're assuming any single json file can fit into memory here because we need to be able to
        # access its internals to extract data from it

        for source_file in source_files:

            try:
                if not self.record_path:
                    df = pd.read_json(source_file, **self.reader_kwargs)
                    df._source_file = source_file
                else:
                    data = self._extract_data(source_file, self.record_path)
                    df = pd.read_json(data, **self.reader_kwargs)
                    df._source_file = source_file

                yield df.transpose() if self.transpose else df
            except Exception as ex:
                if self.skip_errors:
                    print(f'skipping {source_file} due to error: {ex}')
                    yield pd.DataFrame()
                else:
                    raise ex

    def transform(self, source_files: List[Path]) -> pd.DataFrame:

        # TODO: move the renaming logic to the base and allow a mapper to be passed

        for df in self._build_data_frame(source_files):
            if not df.empty:
                if self.rename:
                    df = df.rename(self.rename, axis='columns')
                if self.columns:
                    df = df[self.columns]

            yield df



import pytest
import io
import pandas as pd
import json

from inflection import underscore
from pathlib import Path
from tempfile import NamedTemporaryFile

from ketl.transformer.Transformer import BaseTransformer, DelimitedTableTransformer, JsonTableTransformer


def test_base_transformer():

    transformer = BaseTransformer(**{'foo': 'bar'})

    assert transformer.passed_kwargs == {'foo': 'bar'}

    with pytest.raises(NotImplementedError):
        transformer.transform([Path('foo')])

    with pytest.raises(NotImplementedError):
        transformer._build_data_frame([Path('foo')])


def test_delimited_transformer_init():

    kwargs = {'foo': 'bar', 'transpose': True, 'concat_on_axis': True}
    transformer = DelimitedTableTransformer(**kwargs)

    assert transformer.transpose
    assert transformer.concat_on_axis
    assert transformer.reader_kwargs['foo'] == 'bar'


def test_build_data_frame():

    transformer = DelimitedTableTransformer(delimiter=',')

    data1 = 'x,y,z\n1,2,3\n4,5,6'
    data2 = 'x,y,z\n7,8,9\n10,11,12'
    s1 = io.StringIO()
    s1.write(data1)
    s1.seek(0)
    s2 = io.StringIO()
    s2.write(data2)
    s2.seek(0)

    df1_expected = pd.DataFrame.from_records([(1, 2, 3), (4, 5, 6)], columns=['x', 'y', 'z'])
    df2_expected = pd.DataFrame.from_records([(7, 8, 9), (10, 11, 12)], columns=['x', 'y', 'z'])

    for i, df in enumerate(transformer._build_data_frame([s1, s2])):
        assert len(df) == 2
        if i == 0:
            assert df.equals(df1_expected)
        elif i == 1:
            assert df.equals(df2_expected)
        else:
            assert False

    s1.seek(0)
    s2.seek(0)

    for i, df in enumerate(transformer.transform([s1, s2])):
        assert len(df) == 2
        if i == 0:
            assert df.equals(df1_expected)
        elif i == 1:
            assert df.equals(df2_expected)
        else:
            assert False


def test_build_data_frame_transpose():

    transformer = DelimitedTableTransformer(delimiter=',', transpose=True)

    data1 = 'x,y,z\n1,2,3\n4,5,6'
    data2 = 'x,y,z\n7,8,9\n10,11,12'
    s1 = io.StringIO()
    s1.write(data1)
    s1.seek(0)
    s2 = io.StringIO()
    s2.write(data2)
    s2.seek(0)

    df1_expected = pd.DataFrame.from_records([(1, 4), (2, 5), (3, 6)], columns=[0, 1], index=['x', 'y', 'z'])
    df2_expected = pd.DataFrame.from_records([(7, 10), (8, 11), (9, 12)], columns=[0, 1], index=['x', 'y', 'z'])

    for i, df in enumerate(transformer._build_data_frame([s1, s2])):
        assert len(df) == 3
        if i == 0:
            assert df.equals(df1_expected)
        elif i == 1:
            assert df.equals(df2_expected)
        else:
            assert False

    s1.seek(0)
    s2.seek(0)

    for i, df in enumerate(transformer.transform([s1, s2])):
        assert len(df) == 3
        if i == 0:
            assert df.equals(df1_expected)
        elif i == 1:
            assert df.equals(df2_expected)
        else:
            assert False


def test_build_data_frame_concat():

    transformer = DelimitedTableTransformer(delimiter=',', concat_on_axis=1, iterator=False, chunksize=None)

    data1 = 'x\n1\n2'
    data2 = 'y\n3\n4'
    s1 = io.StringIO()
    s1.write(data1)
    s1.seek(0)
    s2 = io.StringIO()
    s2.write(data2)
    s2.seek(0)

    df1_expected = pd.DataFrame.from_records([(1, 3), (2, 4)], columns=['x', 'y'])
    df2_expected = pd.DataFrame.from_records([(7, 10), (8, 11), (9, 12)], columns=[0, 1], index=['x', 'y', 'z'])

    for i, df in enumerate(transformer._build_data_frame([s1, s2])):
        assert len(df) == 2
        if i == 0:
            assert df.equals(df1_expected)
        else:
            assert False


def test_json_transformer_init():

    kwargs = {'transpose': True, 'concat_on_axis': True, 'record_path': ['path1', 'path2'],
              'rename': underscore, 'columns': ['col1', 'col2']}
    transformer = JsonTableTransformer(**kwargs)

    assert transformer.transpose
    assert transformer.concat_on_axis
    assert transformer.record_path == ['path1', 'path2']
    assert transformer.columns == ['col1', 'col2']
    assert transformer.rename is not None


def test_json_transformer_extract_data(tmp_path):

    data = {
        'foo': {
            'bar': [
                {'field1': 'foo'},
                {'field2': 'bar'}
            ]
        }
    }

    tf = NamedTemporaryFile(dir=tmp_path, delete=False)
    tf.write(json.dumps(data).encode('utf-8'))
    tf.close()

    assert json.dumps(data['foo']['bar']) == JsonTableTransformer._extract_data(tf.name, ['foo', 'bar'])
    assert json.dumps(data['foo']) == JsonTableTransformer._extract_data(tf.name, 'foo')

    with pytest.raises(TypeError):
        JsonTableTransformer._extract_data(tf.name, 1)


def test_json_transformer_build_data_frame(tmp_path):

    df_expected = pd.DataFrame.from_records([('foo1', 'bar1'), ('foo2', 'bar2')], columns=['field1', 'field2'])

    data = [
        {'field1': 'foo1', 'field2': 'bar1'},
        {'field1': 'foo2', 'field2': 'bar2'}
    ]

    tf = NamedTemporaryFile(dir=tmp_path, delete=False)
    tf.write(json.dumps(data).encode('utf-8'))
    tf.close()

    transformer = JsonTableTransformer()

    for df in transformer.transform([tf.name]):
        assert df_expected.equals(df)

    data = {
        'foo': {
            'bar': [
                {'field1': 'foo1', 'field2': 'bar1'},
                {'field1': 'foo2', 'field2': 'bar2'}
            ]
        }
    }

    tf = NamedTemporaryFile(dir=tmp_path, delete=False)
    tf.write(json.dumps(data).encode('utf-8'))
    tf.close()

    transformer = JsonTableTransformer(record_path=['foo', 'bar'])

    for df in transformer.transform([tf.name]):
        assert df_expected.equals(df)

    # mock_record_path = mock.Mock()
    # mock_record_path.side_effect = [ValueError, ValueError]
    #
    # transformer.record_path = mock_record_path


def test_json_transformer_build_df_errors():

    transformer = JsonTableTransformer()

    transformer.skip_errors = True
    dfs = transformer._build_data_frame([Path('file1')])
    for df in dfs:
        assert df.empty


def test_json_transformer_snakecase_and_filter(tmp_path):

    df_expected = pd.DataFrame.from_records([('foo1', ), ('foo2', )], columns=['my_field1'])

    data = [
        {'myField1': 'foo1', 'myField2': 'bar1'},
        {'myField1': 'foo2', 'myField2': 'bar2'}
    ]

    tf = NamedTemporaryFile(dir=tmp_path, delete=False)
    tf.write(json.dumps(data).encode('utf-8'))
    tf.close()

    transformer = JsonTableTransformer(columns=['my_field1'], rename=underscore)

    for df in transformer.transform([tf.name]):
        assert df_expected.equals(df)



import pytest
import zipfile
import gzip
import tarfile
import lzma
import shutil

from tempfile import NamedTemporaryFile, TemporaryDirectory
from pathlib import Path
from hashlib import sha1

from tests.factories import APIFactory, SourceFactory, CachedFileFactory, ExpectedFileFactory, CredsFactory

from ketl.db.settings import get_session
from ketl.db import models


def test_api_config(session):

    api = models.API(name='my nice api')
    api.setup()

    assert api.id is not None

    api = get_session().query(models.API).filter(models.API.name == 'my nice api').one()
    assert api.name == 'my nice api'

    api = models.API()
    api.setup()
    api = get_session().query(models.API).filter(models.API.name == 'API').one()
    assert api.name == 'API'


def test_api_hash(tmp_path):

    api = APIFactory(name='my nice api')
    initial_api_hash = api.api_hash
    expected_api_hash = sha1(b'my nice api')

    assert initial_api_hash == expected_api_hash.hexdigest()

    source: models.Source = SourceFactory(base_url='http://base.url', data_dir='/download/dir', api_config=api)

    with NamedTemporaryFile(dir=tmp_path, delete=False) as tf:
        tf.write(b'hello world - cached file')
        tf.close()

        cached_file: models.CachedFile = CachedFileFactory(source=source, path=tf.name)

    expected_api_hash.update(source.source_hash.digest())
    api_hash = api.api_hash

    assert api_hash != initial_api_hash
    assert api_hash == expected_api_hash.hexdigest()


def test_api_get_instance(session):

    instance1 = models.API.get_instance(models.API)
    assert isinstance(instance1, models.API)
    instance2 = models.API.get_instance(models.API)
    assert isinstance(instance2, models.API)
    assert instance1 == instance2


def test_api_get_expected_files(session):

    api = APIFactory()
    source = SourceFactory(api_config=api)
    cf1 = CachedFileFactory(path='path1', source=source)
    cf2 = CachedFileFactory(path='path2', source=source)
    ef1 = ExpectedFileFactory(path='file1', cached_file=cf1)
    ef2 = ExpectedFileFactory(path='file2', cached_file=cf1)
    ef3 = ExpectedFileFactory(path='file3', cached_file=cf2)

    assert api.expected_files == [ef1, ef2, ef3]
    assert source.expected_files == [ef1, ef2, ef3]


def test_expected_file_hash(session, tmp_path):

    expected_file: models.ExpectedFile = ExpectedFileFactory(path='/path/to/nonexistent/file')
    default_hash = expected_file.file_hash
    assert default_hash is not None

    with NamedTemporaryFile(dir=tmp_path) as tf:
        tf.write(b'hello world')

        expected_file: models.ExpectedFile = ExpectedFileFactory(path=tf.name)
        assert expected_file.file_hash != default_hash
        assert len(expected_file.file_hash.hexdigest()) > 0


def test_cached_file_cache(session, tmp_path):

    missing_cached_file: models.CachedFile = CachedFileFactory(path='/path/to/missing/file')
    assert missing_cached_file.file_hash.hexdigest() == sha1().hexdigest()

    with NamedTemporaryFile(dir=tmp_path, delete=False) as tf:
        tf.write(b'hello world - cached file')
        tf.close()

        cached_file: models.CachedFile = CachedFileFactory(path=tf.name)

        cached_file_hash = cached_file.file_hash

        assert cached_file_hash.hexdigest() != sha1().hexdigest()


def test_source_hash(session, tmp_path):

    source: models.Source = SourceFactory(base_url='http://base.url', data_dir='/download/dir')
    source_hash = source.source_hash.hexdigest()
    assert source_hash == sha1(bytes(source.base_url + source.data_dir, 'utf-8')).hexdigest()

    with NamedTemporaryFile(dir=tmp_path, delete=False) as tf:
        tf.write(b'hello world - cached file')
        tf.close()

        cached_file: models.CachedFile = CachedFileFactory(source=source, path=tf.name)

    expected_hash = sha1(bytes(source.base_url + source.data_dir, 'utf-8'))
    expected_hash.update(cached_file.file_hash.digest())

    source_hash_with_cached_file = source.source_hash.hexdigest()
    assert source_hash_with_cached_file != source_hash
    assert source_hash_with_cached_file == expected_hash.hexdigest()


def test_cached_file_full_path(session):

    source: models.Source = SourceFactory(base_url='http://base.url', data_dir='/download/dir')
    cached_file: models.CachedFile = CachedFileFactory(path='path/to/file', source=source)

    assert cached_file.full_path == Path('/download/dir/path/to/file')


def test_cached_file_extract_gzip(tmp_path):

    tf = NamedTemporaryFile(dir=tmp_path, delete=False)
    tf.write(b'hello world - gzip')
    tf.close()

    raw_file = Path(tf.name)
    compressed_file = Path(tmp_path) / 'test.gz'
    result_file = Path(tmp_path) / 'test'

    with open(raw_file, 'rb') as tf:
        with gzip.open(compressed_file, mode='wb') as cf:
            shutil.copyfileobj(tf, cf)

    cached_file: models.CachedFile = CachedFileFactory(path=str(compressed_file), is_archive=True,
                                                       extract_to=str(tmp_path))
    expected_file: models.ExpectedFile = ExpectedFileFactory(path=str(result_file), cached_file=cached_file)
    cached_file.preprocess()

    assert Path(expected_file.path).exists()

    with open(expected_file.path, 'rb') as f:
        data = f.read()
        assert data == b'hello world - gzip'


def test_cached_file_extract_lzma(tmp_path):

    tf = NamedTemporaryFile(dir=tmp_path, delete=False)
    tf.write(b'hello world - lzma')
    tf.close()

    raw_file = Path(tf.name)
    compressed_file = Path(tmp_path) / 'test.xz'
    result_file = Path(tmp_path) / 'test'

    with open(raw_file, 'rb') as tf:
        with lzma.open(compressed_file, mode='wb') as cf:
            shutil.copyfileobj(tf, cf)

    cached_file: models.CachedFile = CachedFileFactory(path=str(compressed_file), is_archive=True,
                                                       extract_to=str(tmp_path))
    expected_file: models.ExpectedFile = ExpectedFileFactory(path=str(result_file), cached_file=cached_file)
    cached_file.preprocess()

    assert Path(expected_file.path).exists()

    with open(expected_file.path, 'rb') as f:
        data = f.read()
        assert data == b'hello world - lzma'


def test_cached_file_extract_zip(tmp_path):

    temp_path = Path(tmp_path)

    tf1 = NamedTemporaryFile(dir=tmp_path, delete=False)
    tf1.write(b'expected file')
    tf1.close()
    tf1_path = temp_path / Path(tf1.name).name

    tf2 = NamedTemporaryFile(dir=tmp_path, delete=False)
    tf2.write(b'not expected file')
    tf2.close()
    tf2_path = temp_path / Path(tf2.name).name

    compressed_file = Path(tmp_path) / 'test.zip'

    with zipfile.ZipFile(compressed_file, mode='w') as zf:
        zf.write(tf1_path, arcname=tf1_path.name)
        zf.write(tf2_path, arcname=tf2_path.name)

    tf1_path.unlink()
    tf2_path.unlink()

    cached_file: models.CachedFile = CachedFileFactory(path=str(compressed_file), is_archive=True,
                                                       extract_to=str(tmp_path),
                                                       expected_mode=models.ExpectedMode.explicit)
    expected_file: models.ExpectedFile = ExpectedFileFactory(path=str(tf1_path), cached_file=cached_file)

    cached_file.preprocess()

    assert Path(expected_file.path).exists()
    assert not Path(tf2.name).exists()

    with open(expected_file.path, 'rb') as f:
        data = f.read()
        assert data == b'expected file'

    Path(expected_file.path).unlink()

    cached_file.expected_mode = models.ExpectedMode.auto
    cached_file.preprocess()

    assert len(cached_file.expected_files) == 2

    found1 = False
    found2 = False
    for ef in cached_file.expected_files:
        with open(ef.path, 'rb') as f:
            data = f.read()
            if data == b'expected file':
                found1 = True
            elif data == b'not expected file':
                found2 = True

    assert found1 and found2


def test_cached_file_extract_tar(session, tmp_path):

    temp_path = Path(tmp_path)

    tf1 = NamedTemporaryFile(dir=tmp_path, delete=False)
    tf1.write(b'expected file')
    tf1.close()
    tf1_path = temp_path / Path(tf1.name).name

    tf2 = NamedTemporaryFile(dir=tmp_path, delete=False)
    tf2.write(b'not expected file')
    tf2.close()
    tf2_path = temp_path / Path(tf2.name).name

    compressed_file = Path(tmp_path) / 'test.tar.gz'

    with tarfile.open(compressed_file, mode='w:gz') as cf:
        cf.add(tf1_path, arcname=tf1_path.name)
        cf.add(tf2_path, arcname=tf2_path.name)

    tf1_path.unlink()
    tf2_path.unlink()

    cached_file: models.CachedFile = CachedFileFactory(path=str(compressed_file), is_archive=True,
                                                       extract_to=str(tmp_path),
                                                       expected_mode=models.ExpectedMode.explicit)
    expected_file: models.ExpectedFile = ExpectedFileFactory(path=str(tf1_path), cached_file=cached_file)

    cached_file.preprocess()

    assert Path(expected_file.path).exists()
    assert not Path(tf2.name).exists()

    with open(expected_file.path, 'rb') as f:
        data = f.read()
        assert data == b'expected file'

    Path(expected_file.path).unlink()

    cached_file.expected_mode = models.ExpectedMode.auto
    cached_file.preprocess()

    assert len(cached_file.expected_files) == 2

    found1 = False
    found2 = False
    for ef in cached_file.expected_files:
        with open(ef.path, 'rb') as f:
            data = f.read()
            if data == b'expected file':
                found1 = True
            elif data == b'not expected file':
                found2 = True

    assert found1 and found2
